/*
 * dclock messages for Japanese
 *   By: YOKOTA Hiroshi <yokota@netlab.is.tsukuba.ac.jp>
 *   Modified by: YAMAGUCHI Shingo <shingo-y@spacelan.ne.jp>
 */

#ifndef __DCLOCK_MSG_JA_H__
#define __DCLOCK_MSG_JA_H__

#define POPEX_STRING		"現在の世界人口"
#define PEOPLE_STRING		" 人"
#define FOREST_STRING		"現在の熱帯雨林の面積"
#define TROPICAL_STRING		" "
#define HIV_STRING		"現在の HIV 感染数"
#define CASES_STRING		" 件"
#define LAB_STRING		"今年、動物実験により"
#define VEG_STRING		"今年、人間の食糧として"
#define YEAR_STRING		" 体の動物が犠牲になりました"
#define Y2K_STRING		"Y2K (2000年1月1日, 0時00分00秒) まで あと"
#define POST_Y2K_STRING		"Y2K (2000年1月1日, 0時00分00秒) から"
#define Y2001_STRING		"二千年紀の終了 (2001年1月1日, 0時00分00秒) まで あと"
#define POST_Y2001_STRING	"三千年紀の開始 (2001年1月1日, 0時00分00秒) から"
#define MAYAN_STRING		"Countdown to 13th bak'tun (21 December 2012)"
#define POST_MAYAN_STRING	"Countdown to 13th bak'tun (21 December 2012)"
#define DAY			"日"
#define DAYS			"日"
#define HOUR			"時間"
#define HOURS			"時間"
#define MINUTE			"分"
#define MINUTES			"分"
#define SECOND			"秒"
#define SECONDS			"秒"

#ifdef METRIC
#define AREA_STRING		"ヘクタール"
#else
#define AREA_STRING		"エーカー"
#endif

#endif /* __DCLOCK_MSG_JA_H__ */
