/*
 * Resource messages for French
 */

#ifndef __RESOURCE_MSG_FR__
#define __RESOURCE_MSG_FR__

#define DEF_NAME		"Nom : "
#define DEF_PASS		"Mot de passe : "
#define DEF_VALID		"Validation ..."
#define DEF_INVALID		"Mot de passe invalide."
#define DEF_INVALIDCAPSLOCK	"Mot de passe invalide, vérrouillage des majuscules actif."
#define DEF_INFO		"Entrez le mot de passe ou choisissez l'icône pour verrouiller."

#ifdef HAVE_KRB5
#define DEF_KRBINFO		"Entrez le mot de passe Kerberos ou choisissez l'icône pour verrouiller."
#endif /* HAVE_KRB5 */

#define DEF_COUNT_FAILED	" essai a échoué."
#define DEF_COUNTS_FAILED	" essais ont échoués."

#define DEF_BTN_LABEL		"Fermer la session"
#define DEF_BTN_HELP		"Cliquer ici pour fermer la session"
#define DEF_FAIL		"Echec de la fermeture de session automatique"

#endif /* __RESOURCE_MSG_FR__ */
