/*
 * dclock messages for English
 */

#ifndef __DCLOCK_MSG_EN_H__
#define __DCLOCK_MSG_EN_H__

#define POPEX_STRING		"World Population"
#define PEOPLE_STRING		" People"
#define FOREST_STRING		"Tropical Forest Countdown"
#define TROPICAL_STRING		" Tropical Forest in "
#define HIV_STRING		"Current HIV Infections World Wide"
#define CASES_STRING		" Cases"
#define LAB_STRING		"Used in research"
#define VEG_STRING		"Consumed for Food by Mankind"
#define YEAR_STRING		" animals this year"
#define Y2K_STRING		"Countdown to Y2K (1 January 2000, 0:00 hour)"
#define POST_Y2K_STRING		"Time since Y2K (1 January 2000)"
#define Y2001_STRING		"Countdown to the end of the Second Millennium (1 January 2001, 0:00 hour)"
#define POST_Y2001_STRING	"Time since the start of the Third Millennium (1 January 2001)"
#define MAYAN_STRING		"Countdown to 13th bak'tun (21 December 2012)"
#define POST_MAYAN_STRING	"Time since new bak'tun (21 December 2012)"

#define DAY			"day"
#define DAYS			"days"
#define HOUR			"hour"
#define HOURS			"hours"
#define MINUTE			"minute"
#define MINUTES			"minutes"
#define SECOND			"second"
#define SECONDS			"seconds"

#ifdef METRIC
#define AREA_STRING		"Hectares"
#else
#define AREA_STRING		"Acres"
#endif /* METRIC */

#endif /* __DCLOCK_MSG_EN_H__ */
