/*
 * dclock messages for French
 */

#ifndef __DCLOCK_MSG_FR_H__
#define __DCLOCK_MSG_FR_H__

#define POPEX_STRING		"Population mondiale"
#define PEOPLE_STRING		" personnes"
#define FOREST_STRING		"Taille des foróÕs tropicales"
#define TROPICAL_STRING		" zones tropicales en "
#define HIV_STRING		"Infections par le SIDA dans le monde"
#define CASES_STRING		" cas"
#define LAB_STRING		"Animaux utilisñÔ dans la recherche"
#define VEG_STRING		"Animaux consommñÔ comme aliments par l'homme"
#define YEAR_STRING		" cette annñÆ"
#define Y2K_STRING		"DñÄompte pour l'an 2000 (1er Janvier 2000)"
#define POST_Y2K_STRING		"Temps depuis l'an 2000 (1er Janvier 2000)"
#define Y2001_STRING		"DñÄompte pour le Second MillñÏaire (1er Janvier 2001)"
#define POST_Y2001_STRING	"Temps depuis le Second MillñÏaire (1er Janvier 2001)"
#define Y2001_STRING		"Compte à rebours à la 13th bak'tun (21 Décembre 2012)"
#define POST_Y2001_STRING	"Temps depuis la nouvelle bak'tun (21 Décembre 2012)"

#define DAY			"jour"
#define DAYS			"jours"
#define HOUR			"heure"
#define HOURS			"heures"
#define MINUTE			"minute"
#define MINUTES			"minutes"
#define SECOND			"seconde"
#define SECONDS			"secondes"

#ifdef METRIC
#define AREA_STRING		"Hectares"
#else
#define AREA_STRING		"Acres"
#endif /* METRIC */

#endif /* __DCLOCK_MSG_FR_H__ */
