/*
 * XLock messages for French.
 */

#ifndef __XLOCK_MSG_FR_H__ 
#define __XLOCK_MSG_FR_H__

#define TIME_ELAPSED_MINUTE	"%d minute écoulée depuis verrouillage.                      \n"
#define TIME_ELAPSED_MINUTES	"%d minute écoulée depuis verrouillage.                      \n"
#define TIME_ELAPSED_HOURS	"%d:%02d heures écoulées depuis verouillage.                        \n"
#define BUTTON_MINUTE		"%d minute jusqu'à apparition du bouton lougout.             \n"
#define BUTTON_MINUTES		"%d minute%s jusqu'à apparition du bouton lougout.             \n"
#define BUTTON_HOURS		"%d:%02d heures jusqu'à apparition du bouton lougout         \n"
#define AUTOLOGOUT_MINUTE	"%d minute jusqu'à l'auto-logout.    \n"
#define AUTOLOGOUT_MINUTES	"%d minutes jusqu'à l'auto-logout.    \n"
#define AUTOLOGOUT_HOURS	"%d:%02d heure%s jusqu'à auto-logout.  \n"

#endif
