/*
 * dclock messages for Traditional Chinese
 */

#ifndef __DCLOCK_MSG_ZH_TW_H__
#define __DCLOCK_MSG_ZH_TW_H__

#define POPEX_STRING		"世界人口"
#define PEOPLE_STRING		" 人"
#define FOREST_STRING		"熱帶雨林倒數"
#define TROPICAL_STRING		" 熱帶雨林消失於 "
#define HIV_STRING		"目前全世界愛滋病感染數"
#define CASES_STRING		" 例"
#define LAB_STRING		"研究用途"
#define VEG_STRING		"人類所消耗的食物"
#define YEAR_STRING		" 隻動物於今年"
#define Y2K_STRING		"倒數到 Y2K (2000 年 1 月 1 日 0:00)"
#define POST_Y2K_STRING		"時間自 Y2K (2000 年 1 月 1 日)"
#define Y2001_STRING		"倒數到第二千禧年的結束 (2001 年 1 月 1 日 0:00)"
#define POST_Y2001_STRING	"時間自第三千禧年的開始 (2001 年 1 月 1 日)"
#define MAYAN_STRING		"Countdown to 13th bak'tun (21 December 2012)"
#define POST_MAYAN_STRING	"Time since new bak'tun (21 December 2012)"

#define DAY			"天"
#define DAYS			"天"
#define HOUR			"小時"
#define HOURS			"小時"
#define MINUTE			"分"
#define MINUTES			"分"
#define SECOND			"秒"
#define SECONDS			"秒"

#ifdef METRIC
#define AREA_STRING		"公頃"
#else
#define AREA_STRING		"英畝"
#endif /* METRIC */

#endif /* __DCLOCK_MSG_ZH_TW_H__ */
