/*
 * dclock messages for Netherlands language
 */

#ifndef __DCLOCK_MSG_NL_H__
#define __DCLOCK_MSG_NL_H__

#define POPEX_STRING		"Wereld populatie"
#define PEOPLE_STRING		" Mensen"
#define FOREST_STRING		"Aantal tropische wouden"
#define TROPICAL_STRING		" Tropisch gebied in "
#define HIV_STRING		"Huidige staat HIV infecties wereldwijd"
#define CASES_STRING		" gevallen"
#define LAB_STRING		"Verbruikt tijdens onderzoek"
#define VEG_STRING		"Opgegeten door de mens"
#define YEAR_STRING		" dieren dit jaar"
#define Y2K_STRING		"Aftellen tot Y2K (1 Januari 2000, 0.00 uur)"
#define POST_Y2K_STRING		"Tijd sinds Y2K (1 Januari 2000)"
#define Y2001_STRING		"Aftellen tot het einde van het tweede Millennium (1 Januari 2001, 0.00 uur)"
#define POST_Y2001_STRING	"Verstreken tijd sinds het begin van het derde Millennium (1 Januari 2001)"
#define Y2001_STRING		"Aftellen tot het tweede millennium (1 Januari 2001)"
#define POST_Y2001_STRING	"Tijd sinds het tweede millennium (1 Januari 2001)"
#define MAYAN_STRING		"Aftelprocedure aan 13de bak'tun (21 December 2012)"
#define POST_MAYAN_STRING	"Tijd sinds nieuw bak'tun (21 December 2012)"

#define DAY			"dag"
#define DAYS			"dagen"
#define HOUR			"uur"
#define HOURS			"uren"
#define MINUTE			"minuut"
#define MINUTES			"minuten"
#define SECOND			"seconde"
#define SECONDS			"seconden"

#ifdef METRIC
#define AREA_STRING		"Hectare"
#else
#define AREA_STRING		"Acre"
#endif /* METRIC */

#endif /* __DCLOCK_MSG_NL_H__ */
