/*
 * XLock messages for German.
 */

#ifndef __XLOCK_MSG_DE_H__ 
#define __XLOCK_MSG_DE_H__
 
#define TIME_ELAPSED_MINUTE	"Seit %d Minute gesperrt.                                  \n"
#define TIME_ELAPSED_MINUTES	"Seit %d Minuten gesperrt.                                  \n"
#define TIME_ELAPSED_HOURS	"Seit %d:%02d Stunden gesperrt.                                 \n"
#define BUTTON_MINUTE		"In %d Minute erscheint der Auslogger.                       \n"
#define BUTTON_MINUTES		"In %d Minuten erscheint der Auslogger.                       \n"
#define BUTTON_HOURS		"In %d:%02d Stunden erscheint der Auslogger.                 \n"
// Guessing here
#define AUTOLOGOUT_MINUTE	"%d Minute bis Auto-Logout.    \n"
#define AUTOLOGOUT_MINUTES	"%d Minuten bis Auto-Logout.    \n"
#define AUTOLOGOUT_HOURS	"%d:%02d Stunden bis Auto-Logout.  \n"

#endif
