/*
 * XLock messages for English.
 */

#ifndef __XLOCK_MSG_EN_H__ 
#define __XLOCK_MSG_EN_H__
 
#define TIME_ELAPSED_MINUTE	"%d minute elapsed since locked.                           \n"
#define TIME_ELAPSED_MINUTES	"%d minutes elapsed since locked.                           \n"
#define TIME_ELAPSED_HOURS	"%d:%02d hours elapsed since locked.                            \n"
#define BUTTON_MINUTE		"%d minute until the public logout button appears.           \n"
#define BUTTON_MINUTES		"%d minutes until the public logout button appears.           \n"
#define BUTTON_HOURS		"%d:%02d hours until the public logout button appears.       \n"
#define AUTOLOGOUT_MINUTE	"%d minute until auto-logout.    \n"
#define AUTOLOGOUT_MINUTES	"%d minutes until auto-logout.    \n"
#define AUTOLOGOUT_HOURS	"%d:%02d hours until auto-logout.  \n"

#endif
