/*
 * XLock messages for Dutch.
 */

#ifndef __XLOCK_MSG_NL_H__ 
#define __XLOCK_MSG_NL_H__
 
#define TIME_ELAPSED_MINUTE	"%d minuut op slot.                                        \n"
#define TIME_ELAPSED_MINUTES	"%d minuten op slot.                                        \n"
#define TIME_ELAPSED_HOURS	"%d:%02d uur op slot.                                           \n"
#define BUTTON_MINUTE		"Over %d minuut verschijnt de logout knop.                   \n"
#define BUTTON_MINUTES		"Over %d minuten verschijnt de logout knop.                   \n"
#define BUTTON_HOURS		"Over %d:%02d uur verschijnt de logout knop.                 \n"
#define AUTOLOGOUT_MINUTE	"%d minuut tot auto-logout.            \n"
#define AUTOLOGOUT_MINUTES	"%d minuten tot auto-logout.            \n"
#define AUTOLOGOUT_HOURS	"%d:%02d uur tot auto-logout.          \n"
 
#endif
