/*
 * XLock messages for Japanese.
 *   By: YOKOTA Hiroshi <yokota@netlab.is.tsukuba.ac.jp>
 *   Modified by: YAMAGUCHI Shingo <shingo-y@spacelan.ne.jp>
 */

#ifndef __XLOCK_MSG_JA_H__
#define __XLOCK_MSG_JA_H__

#define TIME_ELAPSED_MINUTE	"ロックしてから %d 分が経過しました。                      \n"
#define TIME_ELAPSED_MINUTES	"ロックしてから %d 分が経過しました。                      \n"
#define TIME_ELAPSED_HOURS	"ロックしてから %d 時間 %02d 分経過しました。              \n"
#define BUTTON_MINUTE		"%d 分後にログアウトボタンが出現します。                   \n"
#define BUTTON_MINUTES		"%d 分後にログアウトボタンが出現します。                   \n"
#define BUTTON_HOURS		"%d 時間 %02d 分後にログアウトボタンが出現します。         \n"
#define AUTOLOGOUT_MINUTE	"自動ログアウトまで %d 分"
#define AUTOLOGOUT_MINUTES	"自動ログアウトまで %d 分"
#define AUTOLOGOUT_HOURS	"自動ログアウトまで %d 時間 %02d 分"

#endif
