/*
 * XLock messages for Traditional Chinese.
 *   By: Wei-Lun Chao <chaoweilun@gmail.com>
 */

#ifndef __XLOCK_MSG_ZH_TW_H__
#define __XLOCK_MSG_ZH_TW_H__

#define TIME_ELAPSED_MINUTE	"鎖定之後過了 %d 分鐘。                         \n"
#define TIME_ELAPSED_MINUTES	"鎖定之後過了 %d 分鐘。                         \n"
#define TIME_ELAPSED_HOURS	"鎖定之後過了 %d:%02d 小時。                          \n"
#define BUTTON_MINUTE		"%d 分鐘 直到公開的登出按鈕出現。         \n"
#define BUTTON_MINUTES		"%d 分鐘 直到公開的登出按鈕出現。         \n"
#define BUTTON_HOURS		"%d:%02d 小時直到公用的登出按鈕出現。     \n"
#define AUTOLOGOUT_MINUTE	"%d 分鐘 直到自動登出。  \n"
#define AUTOLOGOUT_MINUTES	"%d 分鐘 直到自動登出。  \n"
#define AUTOLOGOUT_HOURS	"%d:%02d 小時直到自動登出。\n"

#endif
